/*********************************************************************************
 *                                                                               *
 * kizkool30.c                                                                   *
 *                                                                               *
 * This code has been developped by KizKoool and others for Multidec-SE          *
 * and adapted by Keuleu for xawdecode                                           *
 *                                                                               *
 *                                                                               *
 * This program is free software; you can redistribute it and/or                 *
 * modify it under the terms of the GNU General Public License                   *
 * as published by the Free Software Foundation; either version 2                *
 * of the License, or (at your option) any later version.                        *
 *                                                                               *
 *                                                                               *          
 * This program is distributed in the hope that it will be useful,               *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
 * GNU General Public License for more details.                                  *
 *                                                                               *
 *                                                                               *
 * You should have received a copy of the GNU General Public License             *
 * along with this program; if not, write to the Free Software                   *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    *
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html                *
 *                                                                               *
 *                                                                               *
 *********************************************************************************/
#include <stdio.h>
#include "config.h"
#define KIZKOOOL_30
#include "filtre.h"
#include "kizkool.h"
void set_bluefilter_params (int filter, ...);

#ifdef ENABLE_KIZKOOL30

#define DWORD long

// KizKoool extensions
#define ZEILENZAHL           287
static int nUThreshold;
static int nVThreshold;
static int nYUVThreshold;
static int nPixThreshold;
static int iLine;

// Defined in kizkool.c
/*
extern int iUFilterThresholdPCent;
extern int iVFilterThresholdPCent;
extern int iYUVFilterThresholdPCent;
extern int iPixFilterThreshold;
*/
extern int debug;

extern KIZKOOL_PARAM kizkool_param[NB_KIZKOOL_FILTER];

extern int TypeImage;
extern unsigned int CurrentFrame;
extern int CurrentX;
extern int MovieScale;

static unsigned long long nMaske = 0x00ff00ff00ff00ffLL;
static unsigned long long nMaskeUV = 0xFF00FF00FF00FF00LL;
static unsigned long long MasqueToutSaufU = 0xFFFF00FFFFFF00FFLL;    // Garde tout sauf la composante U
static unsigned long long MasqueToutSaufV = 0x00FFFFFF00FFFFFFLL;    // Garde tout sauf la composante V

void
init_kizkool30_filter (void)
{
  if (debug)
    fprintf(stderr, "kizkool30: init filter\n");
  set_bluefilter_params(BLUEFILTER_K30, KIZKOOL_SMUDGE, SMUDGEFILTER_ON, 0);
  set_bluefilter_params(BLUEFILTER_K30, KIZKOOL_UTHRESHOLD, (U_FILTER_THRESHOLD_K30*65535)/100, 0);
  set_bluefilter_params(BLUEFILTER_K30, KIZKOOL_VTHRESHOLD, (V_FILTER_THRESHOLD_K30*65535)/100, 0);
  set_bluefilter_params(BLUEFILTER_K30, KIZKOOL_YUVTHRESHOLD, (YUV_FILTER_THRESHOLD_K30*65535)/100, 0);
  set_bluefilter_params(BLUEFILTER_K30, KIZKOOL_PIXFILTERTHRESHOLD, (PIX_FILTER_THRESHOLD_K30*65535)/100, 0);
  kizkool_param[BLUEFILTER_K30 - 1].nUThreshold = -1;    //irrelevant
  kizkool_param[BLUEFILTER_K30 - 1].nVThreshold = -1;    //irrelevant
  kizkool_param[BLUEFILTER_K30 - 1].nYUVThreshold = -1;    //irrelevant
}

// Filtre anti-bavures v1.2
//
// Ide et code original par 1Konnu
//
// Adaptation  MultiDec, optimisations et rephasage dynamique
// de l'algorithme par dtection du type d'image par KizKoool
//
// Dtection des frames manquantes par KizKoool sur une ide d'E-nek
//

void
FindFrameType30 (DWORD ** pLinesAddress, unsigned short *pSortLines)
{
  DWORD *pQuelle, *pColorQuelle;
  long *pix_sauve1, *pix_sauve2;
  static DWORD ColorBackupType1[32 * 768];
  static DWORD ColorBackupType2[32 * 768];
  int nb_saturees1, nb_saturees2;
  nb_saturees1 = 0;
  nb_saturees2 = 0;


  nUThreshold =
    (int) (kizkool_param[BLUEFILTER_K30 - 1].iUFilterThresholdPCent * 0.01 *
           0xff);
  nVThreshold =
    (int) (kizkool_param[BLUEFILTER_K30 - 1].iVFilterThresholdPCent * 0.01 *
           0xff);
  nYUVThreshold =
    (int) (kizkool_param[BLUEFILTER_K30 - 1].iYUVFilterThresholdPCent * 0.01 *
           0xff);

  // Dtection des frames loupes

  if (NbFramesManquantes () % 2 == 1)
    TypeImage = !(TypeImage);

  // Dtection du type d'image

  for (iLine = ZEILENZAHL - 64; iLine < ZEILENZAHL - 32; iLine++)
    {

      pQuelle = (DWORD *) pLinesAddress[pSortLines[iLine]];
      pColorQuelle = (DWORD *) pLinesAddress[pSortLines[iLine - 1]];
      pix_sauve1 =
        (DWORD *) ColorBackupType1 + CurrentX * (iLine - ZEILENZAHL + 64);
      pix_sauve2 =
        (DWORD *) ColorBackupType2 + CurrentX * (iLine - ZEILENZAHL + 64);

      if ((((iLine % 2) == 0) && TypeImage)
          || (((iLine % 2) != 0) && !TypeImage))
        {
          // ligne paire de l'image de type 1
          // ou ligne impaire de l'image de type 2
          CorrectionMMX (pQuelle, pix_sauve1, pColorQuelle, MasqueToutSaufV);
          CorrectionMMX (pQuelle, pix_sauve2, pColorQuelle, MasqueToutSaufU);
        }
      else
        {
          // ligne impaire de l'image de type 1
          // ou ligne paire de l'image de type 2
          CorrectionMMX (pQuelle, pix_sauve1, pColorQuelle, MasqueToutSaufU);
          CorrectionMMX (pQuelle, pix_sauve2, pColorQuelle, MasqueToutSaufV);
        }

      nb_saturees1 += Secam_UVValue (pix_sauve1, CurrentX / 2);
      nb_saturees2 += Secam_UVValue (pix_sauve2, CurrentX / 2);
    }

  // Test de rephasage

  if (nb_saturees1 > nb_saturees2 + 4)
    TypeImage = !(TypeImage);
}

// Filtre anti-bavures

void
SecamFilter30 (DWORD ** pLinesAddress, unsigned short *pSortLines)
{
  DWORD *pQuelle, *pColorQuelle;

  for (iLine = 1 + MovieScale + 1; iLine < ZEILENZAHL - MovieScale; iLine++)
    {
      //Initialisation de deux pointeurs:
      // pQuelle pointeur sur la ligne source dans l'image dcode:
      pQuelle = (DWORD *) pLinesAddress[pSortLines[iLine]];
      //pColorQuelle pointeur sur la ligne precedente dans l'image dcode:
      pColorQuelle = (DWORD *) pLinesAddress[pSortLines[iLine - 1]];

      if ((((iLine % 2) == 0) && TypeImage)
          || (((iLine % 2) != 0) && !TypeImage))
        {
          // ligne paire de l'image de type 1 ou ligne impaire de l'image de type 2
          CorrectionMMX (pQuelle, pQuelle, pColorQuelle, MasqueToutSaufV);
        }
      else
        {
          // ligne impaire de l'image de type 1 ou ligne paire de l'image de type 2
          CorrectionMMX (pQuelle, pQuelle, pColorQuelle, MasqueToutSaufU);
        }
    }
}

///////////////////////////////////////////////////////////////
// Dtection des frames manquantes par KizKoool              //
// Sur une ide d'exploiter la variable CurrentFrame d'E-nek //
//                                                           //
// Valeur de retour:                                         //
// Nombre de frames non acquises                             //
///////////////////////////////////////////////////////////////
/*
static int
NbFramesManquantes (void)
{
  static int PrecedentFrame = -1;
  int NbFramesLoupees;

  if (CurrentFrame > PrecedentFrame)
    NbFramesLoupees = CurrentFrame - PrecedentFrame - 1;
  else
    NbFramesLoupees = CurrentFrame - PrecedentFrame + 4;

  PrecedentFrame = CurrentFrame;

  return NbFramesLoupees;
}
*/

static int
NbFramesManquantes(void)
{
  static unsigned int PrecedentFrame=0;
  unsigned int NbFramesLoupees;

  NbFramesLoupees = (CurrentFrame - PrecedentFrame)/2 -1;
  PrecedentFrame = CurrentFrame;
  //printf("NbFramesLoupees  = %u\n", NbFramesLoupees);

  return (int)NbFramesLoupees;
}


// Correction des Couleurs UV en MMX par KizKoool
// Premire traduction MMX par E-Nek
// Bas sur le code C de Kizkool sur ide ORIGINALE de 1Konnu
// Paramtres :
// pQuelle      : Source
// pSauve       : Destination 
// pColorQuelle : Rfrence pour reconstituer la couleur
static void
CorrectionMMX (long *pQuelle, long *pSauve, long *pColorQuelle,
               unsigned long long MasqueUV)
{
  unsigned long MasqueUV32, MasqueInverse32;
  unsigned long long MasqueInverse;
  MasqueUV32 = (unsigned long) MasqueUV;
  MasqueInverse32 = ~MasqueUV32;
  MasqueInverse = ~MasqueUV;

  __asm__ (
            /*
               "pushl %%edi                 \n\t"
               "pushl %%esi                 \n\t"
             */
            "movl %0, %%ecx              \n\t"    // CurrentX
            "movl %1, %%edi              \n\t"    // pQuelle
            "movl %2, %%edx              \n\t"    // pSauve
            "movl %3, %%esi              \n\t"    // pColorQuelle
            "movq %4, %%mm4              \n\t"    // MasqueUV
            "movq %5, %%mm5              \n\t"    // MasqueInverse
            ".align 16                   \n\t"
            "CorrectionMMX.PixSui:       \n\t"
            "movq (%%edi), %%mm0         \n\t"
            "movq (%%esi), %%mm1         \n\t" "pand %%mm4,%%mm0            \n\t"    // mm0 = (*pix_courant  & 0xFFFF00FF)    (2x)
            "pand %%mm5,%%mm1            \n\t"    // mm1 = (*pColorQuelle & 0x0000FF00)
            "por %%mm1,%%mm0             \n\t"    // ((*pix_courant & 0xFFFF00FF) | (*pColorQuelle & 0x0000FF00))
            "movq %%mm0, (%%edx)         \n\t"
            "addl $8,%%edi               \n\t"
            "addl $8,%%edx               \n\t"
            "addl $8,%%esi               \n\t"
            "subl $4,%%ecx               \n\t"
            "jz CorrectionMMX.fin        \n\t"
            "cmpl $2,%%ecx               \n\t"
            "je CorrectionMMX.PixFin     \n\t"
            "jmp CorrectionMMX.PixSui    \n\t" "CorrectionMMX.PixFin:       \n\t"    // Le pixel de fin si ligne de taille impaire
            "movl (%%edi), %%eax         \n\t"
            "movl (%%esi), %%ebx         \n\t" "andl %6, %%eax              \n\t"    // MasqueUV32
            "andl %7, %%ebx              \n\t"    // MasqueInverse32
            "orl %%ebx,%%eax             \n\t"
            "movl %%eax, (%%edx)         \n\t"
            "CorrectionMMX.fin:          \n\t"
            "emms                        \n\t"
            /*
               "popl %%esi                  \n\t"
               "popl %%edi                  \n\t"
             */
:                                /* output regs/vars */
:                                /* input regs */
            "m" (CurrentX),        /* %0 */
            "m" (pQuelle),        /* %1 */
            "m" (pSauve),        /* %2 */
            "m" (pColorQuelle),    /* %3 */
            "m" (MasqueUV),        /* %4 */
            "m" (MasqueInverse),    /* %5 */
            "m" (MasqueUV32),    /* %6 */
            "m" (MasqueInverse32)    /* %7 */
:                                /* clobber list */
            "eax", "ebx", "ecx", "edx", "edi", "esi");
}


/*
void kizkoool_version(void)
{
  fprintf(stderr, "Kizkoool Blue Line Filter 3.0 with anti-smudge 1.2 filter\n\n");
}
*/

// Filtre KizKoool v3.0

// Premires fonctions de traitement SECAM par Fantasio
// Ancien filtre bleu v2.0 + traduction ASM par 1konnu (ide originale de Z)
// Quelques bugfixes par e-nek (e-nek@netcourrier.com)
//
// Nouvel algorithme de dtection des lignes satures + bugfixes +
// harmonisation de ce filtre avec le filtre anti-bavures
// traduction MMX du filtre bleu v2.0 par KizKoool


void
FilterBlueLines30 (DWORD ** pLinesAddress, unsigned short *pSortLines)
{
  int saturee;

  // La zone est limite  287 x 768 (est-ce toujours vrai ?)
  static DWORD ColorBackup[ZEILENZAHL * 768];

  DWORD *pQuelle, *pColorQuelle;

  nUThreshold =
    (int) (kizkool_param[BLUEFILTER_K30 - 1].iUFilterThresholdPCent * 0.01 *
           0xff);
  nVThreshold =
    (int) (kizkool_param[BLUEFILTER_K30 - 1].iVFilterThresholdPCent * 0.01 *
           0xff);
  nYUVThreshold =
    (int) (kizkool_param[BLUEFILTER_K30 - 1].iYUVFilterThresholdPCent * 0.01 *
           0xff);
  nPixThreshold =
    (int) ((CurrentX * kizkool_param[BLUEFILTER_K30 - 1].iPixFilterThreshold)
           / 200);

  for (iLine = 1 + MovieScale; iLine < ZEILENZAHL - MovieScale; iLine++)
    {
      pQuelle = (DWORD *) pLinesAddress[pSortLines[iLine]];
      pColorQuelle = (DWORD *) ColorBackup + CurrentX * iLine;

      saturee = Secam_UVValue (pQuelle, CurrentX / 2);

      ////////////////////////////////////////////////////////////////////////
      // Filtre bleu v2.0
      //
      // Bas sur le code d'1Konnu
      // Corrections du code d'1Konnu par E-Nek
      //
      // Modifications additionnelles de ce filtre par KizKoool:
      //   - Support de la nouvelle interface de la fonction Secam_UVValue
      //   - Traduction du code x86 en code MMX + Optimisations
      ////////////////////////////////////////////////////////////////////////

      __asm__ (
                /*
                   "pushl %%edi                  \n\t"
                   "pushl %%esi                  \n\t"
                 */
                "movl %0, %%ecx               \n\t"    // CurrentX
                "movl %1, %%edi               \n\t"    // pQuelle
                "movl %2, %%esi               \n\t"    // pColorQuelle
                "cmpl $1, %3                  \n\t"    // saturee
                "jge FilterBlueLines.bugne    \n\t"    // saut si la ligne est sature
                //
                // Sauvegarder la ligne en mmoire si elle est bonne
                //
                ".align 16                    \n\t"
                "FilterBlueLines.PixSui:      \n\t"
                "movq (%%edi), %%mm0          \n\t"
                "movq %%mm0, (%%esi)          \n\t"
                "addl $8, %%edi               \n\t"
                "addl $8, %%esi               \n\t"
                "subl $4, %%ecx               \n\t"
                "jz FilterBlueLines.fin       \n\t"
                "cmpl $2, %%ecx               \n\t"
                "je FilterBlueLines.PixFin    \n\t"
                "jmp FilterBlueLines.PixSui   \n\t"
                "FilterBlueLines.PixFin:      \n\t"
                "movl (%%edi), %%eax          \n\t"
                "movl %%eax, (%%esi)          \n\t"
                "jmp FilterBlueLines.fin      \n\t"
                //
                // Correction de la ligne si elle est sature
                //
                "FilterBlueLines.bugne:       \n\t"
                "movq nMaske, %%mm4           \n\t"
                "movq nMaskeUV, %%mm5         \n\t"
                ".align 16                    \n\t"
                "FilterBlueLines.PixSuiv:     \n\t" "movq (%%edi), %%mm0          \n\t"    // mm0 = V2 Y3 U2 Y2 V0 Y1 U0 Y0
                "movq (%%esi), %%mm1          \n\t"    // mm1 = V2 Y3 U2 Y2 V0 Y1 U0 Y0
                "pand %%mm4, %%mm0            \n\t"    // mm0 = 00 Y3 00 Y2 00 Y1 00 Y0
                "pand %%mm5, %%mm1            \n\t"    // mm1 = V2 00 U2 00 V0 00 U0 00
                "por  %%mm1, %%mm0            \n\t"
                "movq %%mm0, (%%edi)          \n\t"
                "addl $8, %%edi               \n\t"
                "addl $8, %%esi               \n\t"
                "subl $4, %%ecx               \n\t"
                "jz FilterBlueLines.fin       \n\t"
                "cmpl $2, %%ecx               \n\t"
                "je FilterBlueLines.PixFinal  \n\t"
                "jmp FilterBlueLines.PixSuiv  \n\t"
                "FilterBlueLines.PixFinal:    \n\t"
                "movl (%%edi), %%eax          \n\t"
                "movl (%%esi), %%ebx          \n\t"
                "andl $0x00ff00ff, %%eax      \n\t"
                "andl $0xff00ff00, %%ebx      \n\t"
                "orl %%ebx, %%eax             \n\t"
                "movl %%eax, (%%edi)          \n\t"
                "FilterBlueLines.fin:         \n\t"
                "emms                         \n\t"
                /*
                   "popl %%esi                   \n\t"
                   "popl %%edi                   \n\t"
                 */
    :                            /* output regs/vars */

    :                            /* input regs */
                "m" (CurrentX),    /* %0 */
                "m" (pQuelle),    /* %1 */
                "m" (pColorQuelle),    /* %2 */
                "m" (saturee)    /* %3 */
    :                            /* clobber list */
                "eax", "ebx", "ecx", "edi", "esi");
    }                            // boucle for
}

////////////////////////////////////////////////////
// Code de la fonction Secam_UVValue par KizKoool //
////////////////////////////////////////////////////

static int
Secam_UVValue (long *pQuelle, int CountX)
{
  long *pix_courant;
  long U, V, Y, Seuil_NbPix;
  int etat_ligne;

  Seuil_NbPix = nPixThreshold;
  pix_courant = pQuelle;

  while ((pix_courant < pQuelle + CountX) && (Seuil_NbPix > 0))
    {
      U = *pix_courant >> 8 & 0xff;
      V = *pix_courant >> 24 & 0xff;
      Y = ((*pix_courant >> 16 & 0xff) + (*pix_courant & 0xff)) >> 1;

      if (!
          ((((U >> 1) + V) > (Y >> 1) + nYUVThreshold)
           && ((U > nUThreshold) || (V > nVThreshold))))
        {
          Seuil_NbPix -= 1;
          // Debug mode
          // *pix_courant=0;
        }
      pix_courant++;
    }

  etat_ligne = (Seuil_NbPix > 0);

  return etat_ligne;
}
#endif
