/*********************************************************************************
 *                                                                               *
 * kizkool31.c                                                                   *
 *                                                                               *
 * This code has been developped by KizKoool and others for Multidec-SE          *
 * and adapted by Keuleu for xawdecode                                           *
 *                                                                               *
 *                                                                               *
 * This program is free software; you can redistribute it and/or                 *
 * modify it under the terms of the GNU General Public License                   *
 * as published by the Free Software Foundation; either version 2                *
 * of the License, or (at your option) any later version.                        *
 *                                                                               *
 *                                                                               *          
 * This program is distributed in the hope that it will be useful,               *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
 * GNU General Public License for more details.                                  *
 *                                                                               *
 *                                                                               *
 * You should have received a copy of the GNU General Public License             *
 * along with this program; if not, write to the Free Software                   *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    *
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html                *
 *                                                                               *
 *                                                                               *
 *********************************************************************************/
#include <stdio.h>
#include "config.h"
#define KIZKOOOL_31
#include "filtre.h"
#include "kizkool.h"
void set_bluefilter_params (int filter, ...);

#ifdef ENABLE_KIZKOOL31

#define DWORD long

// KizKoool extensions
#define ZEILENZAHL           287

static int nUThreshold;
static int nVThreshold;
static int nYUVThreshold;
static int nPixThreshold;
static int iLine;

// Defined in kizkool.c
/*
extern int iUFilterThresholdPCent;
extern int iVFilterThresholdPCent;
extern int iYUVFilterThresholdPCent;
extern int iPixFilterThreshold;
*/
extern int debug;

extern KIZKOOL_PARAM kizkool_param[NB_KIZKOOL_FILTER];

extern int TypeImage;
extern unsigned int CurrentFrame;
extern int CurrentX;
extern int MovieScale;

void
init_kizkool31_filter (void)
{
  if (debug)
    fprintf(stderr, "kizkool31: init filter\n");
  set_bluefilter_params(BLUEFILTER_K31, KIZKOOL_SMUDGE, SMUDGEFILTER_ON, 0);
  set_bluefilter_params(BLUEFILTER_K31, KIZKOOL_UTHRESHOLD, (U_FILTER_THRESHOLD_K31*65535)/100, 0);
  set_bluefilter_params(BLUEFILTER_K31, KIZKOOL_VTHRESHOLD, (V_FILTER_THRESHOLD_K31*65535)/100, 0);
  set_bluefilter_params(BLUEFILTER_K31, KIZKOOL_YUVTHRESHOLD, (YUV_FILTER_THRESHOLD_K31*65535)/100, 0);
  set_bluefilter_params(BLUEFILTER_K31, KIZKOOL_PIXFILTERTHRESHOLD, (PIX_FILTER_THRESHOLD_K31*65535)/100, 0);
  kizkool_param[BLUEFILTER_K31 - 1].nUThreshold = -1;    //irrelevant
  kizkool_param[BLUEFILTER_K31 - 1].nVThreshold = -1;    //irrelevant
  kizkool_param[BLUEFILTER_K31 - 1].nYUVThreshold = -1;    //irrelevant

}

//
// Filtre anti-bavures v1.3
//
// Ide et code original par 1Konnu
//
// Adaptation  MultiDec, optimisations et rephasage dynamique
// de l'algorithme par dtection du type d'image par KizKoool
//
// Dtection des frames manquantes par KizKoool sur une ide d'E-nek
//

void
FindFrameType31 (DWORD ** pLinesAddress, unsigned short *pSortLines)
{
  DWORD *pQuelle, *pColorQuelle;
  long *pLineBackup;
  static DWORD LineColorBackup[768];
  unsigned long MasqueToutSaufU, MasqueToutSaufV;
  int nb_saturees1, nb_saturees2;
  nb_saturees1 = 0;
  nb_saturees2 = 0;

  MasqueToutSaufU = 0xFFFF00FF;    // Garde tout sauf la composante U
  MasqueToutSaufV = 0x00FFFFFF;    // Garde tout sauf la composante V

  pLineBackup = (DWORD *) LineColorBackup;

  nUThreshold =
    (int) (kizkool_param[BLUEFILTER_K31 - 1].iUFilterThresholdPCent * 0.01 *
           0xff);
  nVThreshold =
    (int) (kizkool_param[BLUEFILTER_K31 - 1].iVFilterThresholdPCent * 0.01 *
           0xff);
  nYUVThreshold =
    (int) (kizkool_param[BLUEFILTER_K31 - 1].iYUVFilterThresholdPCent * 0.01 *
           0xff);
  nPixThreshold =
    (int) ((CurrentX * kizkool_param[BLUEFILTER_K31 - 1].iPixFilterThreshold)
           / 200);

  // Dtection des frames loupes

  if (NbFramesManquantes () % 2 == 1)
    TypeImage = !(TypeImage);

  // Dtection du type d'image

  for (iLine = ZEILENZAHL - 64; iLine < ZEILENZAHL - 32; iLine++)
    {

      pQuelle = (DWORD *) pLinesAddress[pSortLines[iLine]];
      pColorQuelle = (DWORD *) pLinesAddress[pSortLines[iLine - 1]];

      if ((((iLine % 2) == 0) && TypeImage)
          || (((iLine % 2) != 0) && !TypeImage))
        {
          // ligne paire de l'image de type 1
          // ou ligne impaire de l'image de type 2
          nb_saturees1 +=
            TestCorrectionUV (pQuelle, pLineBackup, pColorQuelle,
                              MasqueToutSaufV);
          nb_saturees2 +=
            TestCorrectionUV (pQuelle, pLineBackup, pColorQuelle,
                              MasqueToutSaufU);
        }
      else
        {
          // ligne impaire de l'image de type 1
          // ou ligne paire de l'image de type 2
          nb_saturees1 +=
            TestCorrectionUV (pQuelle, pLineBackup, pColorQuelle,
                              MasqueToutSaufU);
          nb_saturees2 +=
            TestCorrectionUV (pQuelle, pLineBackup, pColorQuelle,
                              MasqueToutSaufV);
        }
    }

  // Test de rephasage

  if (nb_saturees1 > nb_saturees2 + 4)
    TypeImage = !(TypeImage);
}

//
// Dtection des frames manquantes par KizKoool
// Sur une ide d'exploiter la variable CurrentFrame d'E-nek
//
// Valeur de retour:
// Nombre de frames non acquises
//
/*
static int
NbFramesManquantes (void)
{
  static int PrecedentFrame = -1;
  int NbFramesLoupees;

  if (CurrentFrame > PrecedentFrame)
    NbFramesLoupees = CurrentFrame - PrecedentFrame - 1;
  else
    NbFramesLoupees = CurrentFrame - PrecedentFrame + 4;

  PrecedentFrame = CurrentFrame;

  return NbFramesLoupees;
}
*/
static int
NbFramesManquantes(void)
{
  static unsigned int PrecedentFrame=0;
  unsigned int NbFramesLoupees;

  NbFramesLoupees = (CurrentFrame - PrecedentFrame)/2 -1;
  PrecedentFrame = CurrentFrame;
  //printf("NbFramesLoupees  = %u\n", NbFramesLoupees);

  return (int)NbFramesLoupees;
}

//
// Test de correction des Couleurs UV par KizKoool
//
// Paramtres :
// pQuelle      : Source
// pSauve       : Destination 
// pColorQuelle : Rfrence pour reconstituer la couleur
// MasqueUV     : Masque de correction
// 
static int
TestCorrectionUV (long *pQuelle, long *pSauve, long *pColorQuelle,
                  unsigned long MasqueUV)
{
  unsigned long MasqueInverse;
  long *pix_courant;
  MasqueInverse = ~MasqueUV;

  for (pix_courant = pSauve; pix_courant < pSauve + (CurrentX >> 1);
       pix_courant++)
    {
      *pix_courant =
        ((*pQuelle & MasqueUV) | (*pColorQuelle & MasqueInverse));
      pColorQuelle++;
      pQuelle++;
    }
  return LineSaturation (pSauve, CurrentX >> 1);
}

//
// Filtre anti-bavures d'1Konnu
// Adapt  MultiDec et optimis par KizKoool
//
void
SecamFilter31 (DWORD ** pLinesAddress, unsigned short *pSortLines)
{
  BYTE *pQuelle, *pColorQuelle, *pix_courant;

  for (iLine = 1 + MovieScale + 1; iLine < ZEILENZAHL - MovieScale; iLine++)
    {
      // pQuelle: pointeur sur la ligne source dans l'image dcode:
      pQuelle = (BYTE *) pLinesAddress[pSortLines[iLine]];
      // pColorQuelle: pointeur sur la ligne precedente dans l'image dcode:
      pColorQuelle = (BYTE *) pLinesAddress[pSortLines[iLine - 1]];

      if ((((iLine % 2) == 0) && TypeImage)
          || (((iLine % 2) != 0) && !TypeImage))
        {
          // ligne paire de l'image de type 1 ou ligne impaire de l'image de type 2
          pQuelle = pQuelle + 3;
          for (pix_courant = pColorQuelle + 3;
               pix_courant < pColorQuelle + (CurrentX << 1);
               pix_courant = pix_courant + 4)
            {
              *pQuelle = *pix_courant;
              pQuelle = pQuelle + 4;
            }
        }
      else
        {
          // ligne impaire de l'image de type 1 ou ligne paire de l'image de type 2
          pQuelle++;
          for (pix_courant = pColorQuelle + 1;
               pix_courant < pColorQuelle + (CurrentX << 1);
               pix_courant = pix_courant + 4)
            {
              *pQuelle = *pix_courant;
              pQuelle = pQuelle + 4;
            }
        }
    }
}


//
// Filtre KizKoool v3.1
//
// Premires fonctions de traitement SECAM par Fantasio
// Ancien filtre bleu ASM par 1konnu
// Quelques bugfixes par E-nek
//
// Nouvel algorithme de dtection des lignes satures + bugfixes +
// harmonisation de ce filtre avec le filtre anti-bavures +
// traduction en C du filtre bleu et optimisations par KizKoool
//

/*
void FilterBlueLines31_original(DWORD **pLinesAddress, unsigned short *pSortLines)
{
    // Taille maximale d'une frame 768*287
    static BYTE ColorBackup[768 * ZEILENZAHL];

    BYTE *pColorQuelle, *pQuelle, *pix_courant;    

    nUThreshold= (int) (iUFilterThresholdPCent*0.01*0xff);
    nVThreshold= (int) (iVFilterThresholdPCent*0.01*0xff);
    nYUVThreshold= (int) (iYUVFilterThresholdPCent*0.01*0xff);
    nPixThreshold= (int) ((CurrentX*iPixFilterThreshold)/200);
    
    for (iLine=1+MovieScale; iLine < ZEILENZAHL-MovieScale; iLine++)
    {                
        pQuelle      = (BYTE*) pLinesAddress[pSortLines[iLine]];
        pColorQuelle = (BYTE*) (ColorBackup + CurrentX * iLine);

        if (LineSaturation((DWORD*) pQuelle, CurrentX>>1))
        {       
            // Ligne saturee
            // Restauration de la saturation  partir du buffer de lignes non satures
            for (pix_courant=pQuelle+1; pix_courant<pQuelle+(CurrentX<<1); pix_courant=pix_courant+4)
            {
                *pix_courant=*pColorQuelle;
                *(pix_courant+2)=*(pColorQuelle+1);
                pColorQuelle=pColorQuelle+2;
            }
        }
        else
        {       
            // Ligne non saturee
            // Stockage de la ligne dans le buffer de lignes non satures
            for (pix_courant=pQuelle+1; pix_courant<pQuelle+(CurrentX<<1); pix_courant=pix_courant+4)
            {
                *pColorQuelle= *pix_courant;
                *(pColorQuelle+1)= *(pix_courant+2);
                pColorQuelle=pColorQuelle+2;
            }
        }
    } // boucle for
}
*/

// Amelioration par Quenotte
void
FilterBlueLines31 (DWORD ** pLinesAddress, unsigned short *pSortLines)
{
  // Taille maximale d'une frame 768*287
  static BYTE ColorBackup[768 * ZEILENZAHL];
  int Saturation;
  BYTE *pColorQuelle, *pQuelle, *pix_courant;

  nUThreshold =
    (int) (kizkool_param[BLUEFILTER_K31 - 1].iUFilterThresholdPCent * 0.01 *
           0xff);
  nVThreshold =
    (int) (kizkool_param[BLUEFILTER_K31 - 1].iVFilterThresholdPCent * 0.01 *
           0xff);
  nYUVThreshold =
    (int) (kizkool_param[BLUEFILTER_K31 - 1].iYUVFilterThresholdPCent * 0.01 *
           0xff);
  nPixThreshold =
    (int) ((CurrentX * kizkool_param[BLUEFILTER_K31 - 1].iPixFilterThreshold)
           / 200);

  for (iLine = 1 + MovieScale; iLine < ZEILENZAHL - MovieScale; iLine++)
    {
      //Test de saturation de la ligne
      Saturation =
        LineSaturation (pLinesAddress[pSortLines[iLine]], CurrentX >> 1);

      //
      // Filtre anti-bavures d'1Konnu
      // Adapt et optimis par KizKoool
      // Adaptation a K!TV et optimisation a la sauce Quenotte
      //
      if (iLine != 1 + MovieScale)
        {
          // pQuelle: pointeur sur la ligne source dans l'image dcode:
          pQuelle = (BYTE *) pLinesAddress[pSortLines[iLine]];
          // pColorQuelle: pointeur sur la ligne precedente dans l'image dcode:
          pColorQuelle = (BYTE *) pLinesAddress[pSortLines[iLine - 1]];

          if ((((iLine % 2) == 0) && TypeImage)
              || (((iLine % 2) != 0) && !TypeImage))
            {
              // ligne paire de l'image de type 1 ou ligne impaire de l'image de type 2
              pQuelle = pQuelle + 3;
              for (pix_courant = pColorQuelle + 3;
                   pix_courant < pColorQuelle + (CurrentX << 1);
                   pix_courant = pix_courant + 4)
                {
                  *pQuelle = *pix_courant;
                  pQuelle = pQuelle + 4;
                }
            }
          else
            {
              // ligne impaire de l'image de type 1 ou ligne paire de l'image de type 2
              pQuelle++;
              for (pix_courant = pColorQuelle + 1;
                   pix_courant < pColorQuelle + (CurrentX << 1);
                   pix_courant = pix_courant + 4)
                {
                  *pQuelle = *pix_courant;
                  pQuelle = pQuelle + 4;
                }
            }
        }

      pQuelle = (BYTE *) pLinesAddress[pSortLines[iLine]];
      pColorQuelle = (BYTE *) (ColorBackup + CurrentX * iLine);

      //Correction si saturation
      if (Saturation)
        {
          // Ligne saturee
          // Restauration de la saturation  partir du buffer de lignes non satures
          for (pix_courant = pQuelle + 1;
               pix_courant < pQuelle + (CurrentX << 1);
               pix_courant = pix_courant + 4)
            {
              *pix_courant = *pColorQuelle;
              *(pix_courant + 2) = *(pColorQuelle + 1);
              pColorQuelle = pColorQuelle + 2;
            }
        }
      else
        {
          // Ligne non saturee
          // Stockage de la ligne dans le buffer de lignes non satures
          for (pix_courant = pQuelle + 1;
               pix_courant < pQuelle + (CurrentX << 1);
               pix_courant = pix_courant + 4)
            {
              *pColorQuelle = *pix_courant;
              *(pColorQuelle + 1) = *(pix_courant + 2);
              pColorQuelle = pColorQuelle + 2;
            }
        }
    }                            // boucle for
}

//
// Test de saturation de ligne par KizKoool
//
// Paramtres:
// pQuelle: Pointeur vers la ligne  tester
// CountX: Nombre de pixels  tester
//
static int
LineSaturation (long *pQuelle, int CountX)
{
  long *pix_courant;
  long U, V, Y, Seuil_NbPix;

  Seuil_NbPix = nPixThreshold;
  pix_courant = pQuelle;

  while ((pix_courant < pQuelle + CountX) && (Seuil_NbPix > 0))
    {
      U = *pix_courant >> 8 & 0xff;
      V = *pix_courant >> 24 & 0xff;
      Y = ((*pix_courant >> 16 & 0xff) + (*pix_courant & 0xff)) >> 2;

      if (!
          ((((U >> 1) + V) > Y + nYUVThreshold)
           && ((U > nUThreshold) || (V > nVThreshold))))
        {
          Seuil_NbPix -= 1;
        }
      pix_courant++;
    }

  return (Seuil_NbPix > 0);
}
#endif
