/**************************************/
/*                                    */
/*     Nagravision sound decoder      */
/*                                    */
/*          by  Galtor 1999           */
/*                                    */
/*         and by Keuleu 2000         */
/*                                    */
/*         MMX by Maupatz  2001       */
/**************************************/
/* MMX for do_audiodecode_44          */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <sys/time.h>

/* ALSA native support is disabled!!! Please, fix me ! */
#ifdef ALSA
# ifdef HAVE_SYS_ASOUNDLIB_H
#   include <sys/asoundlib.h>
# elif HAVE_ALSA_ASOUNDLIB_H
#   include <alsa/asoundlib.h>
# endif
# include <linux/asound.h>
# include <linux/asoundid.h>
#else
# include <linux/soundcard.h>
# include <sys/ioctl.h>
#endif

#include <pthread.h>

#include "config.h"
#include "nagraudio.h"
#ifdef WRITE_TO_WAV
#  include "wavhead.h"
#endif

#include "attributes.h"
#include "cpu_accel.h"
#include "memcpy.h"
#if defined(HAVE_DIVX4LINUX) || defined (HAVE_FFMPEG) || defined (HAVE_XVID)
# include "divx.h"
#endif

extern int debug;
extern int have_MMX;
extern double audiostamp;

#undef WRITE_TO_WAV

/*******************************************/
/************ FILTER DEFINITION ************/
/*******************************************/

/* The user level n are internal n*2-1 level !! */


static int in_coef[MAX_ORDER + 1][MAX_ORDER] = {
  {}, {}, {}, {0, 3}, {}, {0, 5, 10}, {}, {0, 7, 21, 35}
};

/************ POST MODULATION FILTER DEFINITION **************/

static int post_out_coef[MAX_ORDER + 1][MAX_ORDER + 1] = {
  {},

  {0, 0.0731489272 * (1 << PRECISSION)},    //Level 1 

  {},

  {0, 0.2678654389 * (1 << PRECISSION), -0.3523244053 * (1 << PRECISSION),    //Level 3 
   0.0247302699 * (1 << PRECISSION)},

  {},

  {0, 0.4576942144 * (1 << PRECISSION), -0.7067341465 * (1 << PRECISSION),    //Level 5 
   0.1603017017 * (1 << PRECISSION), -0.0670410490 * (1 << PRECISSION),
   0.0043118519 * (1 << PRECISSION)},

  {},

  {0, 0.6455546255 * (1 << PRECISSION), -1.0827483976 * (1 << PRECISSION),    //Level 7
   0.4131654344 * (1 << PRECISSION), -0.2552919195 * (1 << PRECISSION),
   0.0500299877 * (1 << PRECISSION), -0.0111237227 * (1 << PRECISSION),
   0.0006362623 * (1 << PRECISSION)}
};


static int post_gains[MAX_ORDER + 1] = { 0,
  2.157843972 * (1 << PRECISSION), 0,
  7.549101979 * (1 << PRECISSION), 0,
  27.79062545 * (1 << PRECISSION), 0,
  103.2443130 * (1 << PRECISSION)
};


/************ PRE MODULATION FILTER DEFINITION ***************/

static int pre_out_coef[MAX_ORDER + 1][MAX_ORDER + 1] = {
  {},

  {0, -0.1416306738 * (1 << PRECISSION)},    //Level 1 

  {},

  {0, -0.5168039063 * (1 << PRECISSION), -0.4041813836 * (1 << PRECISSION),    //Level 3
   -0.0497187889 * (1 << PRECISSION)},

  {},

  {0, -0.8823852003 * (1 << PRECISSION), -0.9062925254 * (1 << PRECISSION),    //Level 5
   -0.3366224025 * (1 << PRECISSION), -0.0995725400 * (1 << PRECISSION),
   -0.0096268552 * (1 << PRECISSION)},

  {},

  {0, -1.2442434093 * (1 << PRECISSION), -1.5253656891 * (1 << PRECISSION),    //Level 7
   -0.9065143396 * (1 << PRECISSION), -0.4407574247 * (1 << PRECISSION),
   -0.1209858632 * (1 << PRECISSION), -0.0220286561 * (1 << PRECISSION),
   -0.0016585133 * (1 << PRECISSION)}
};

static int pre_gains[MAX_ORDER + 1] = { 0,
  1.751880048 * (1 << PRECISSION), 0,
  4.059462852 * (1 << PRECISSION), 0,
  9.893338913 * (1 << PRECISSION), 0,
  24.32741402 * (1 << PRECISSION)
};

/*********************************************/
/********** END OF FILTER DEFINITION *********/
/*********************************************/

/************** GLOBAL VARIABLES *************/

static int bits = 16;
static int DEC_FREQ = 44100;
static int OUT_GAIN = 0;        /* 0= no gain, 1= x2, 2= x4 ,3= x8 */

#ifdef ALSA
static snd_pcm_t *handle = NULL;    /* Alsa card handle */
#else
static char device[20] = "/dev/dsp";
static int fd = -1;
#endif

static int decode = 1;
static int auto_decode = 0;
static int pre_filter_level = 0;    /* 0, 1, 3, 5, 7 */
static int post_filter_level = 7;    /* 0, 1, 3, 5, 7 */
static pthread_t nagraudio_thread;
#if defined(HAVE_DIVX4LINUX) || defined (HAVE_FFMPEG)  || defined (HAVE_XVID)
static pthread_mutex_t divx_mutex;
#endif

/*********************************************/

void
start_nagraudio (void)
{
  int arg = 0;                    /* do not read config file, already done by xawdecode */
  int ret;

#if defined(HAVE_DIVX4LINUX) || defined (HAVE_FFMPEG)  || defined (HAVE_XVID)
  pthread_mutex_init (&divx_mutex, NULL);
  if (is_divx_in_progress())
    divx_audio_stop();
#endif
  ret = pthread_create (&nagraudio_thread, NULL, nagraudio_main, (void *) arg);
  if (debug)
    {
      if (ret == 0)
        printf ("Nagraudio started.\n");
      else
        printf ("Nagraudio start failed.\n");
    }
}

void
stop_nagraudio (void)
{
  void *ret;

  pthread_cancel (nagraudio_thread);
  pthread_join (nagraudio_thread, &ret);
  close_card ();
  if (debug)
    printf ("Nagraudio stopped.\n");

#if defined(HAVE_DIVX4LINUX) || defined (HAVE_FFMPEG)  || defined (HAVE_XVID)
  if (is_divx_in_progress())
    divx_audio_start();
  pthread_mutex_destroy(&divx_mutex);
#endif
}

#ifdef WRITE_TO_WAV
void
write_wav_header (int sfdes, int data_length)
{
  wavhead header;

  char *riff = "RIFF";
  char *wave = "WAVE";
  char *fmt = "fmt ";
  char *data = "data";

  fast_memcpy (&(header.main_chunk), riff, 4);
  header.length = sizeof (wavhead) - 8 + data_length;
  fast_memcpy (&(header.chunk_type), wave, 4);

  fast_memcpy (&(header.sub_chunk), fmt, 4);
  header.sc_len = 16;
  header.format = 1;
#ifdef ALSA
  header.modus = MONO_STEREO;
#else
  header.modus = MONO_STEREO + 1;
#endif
  header.sample_fq = DEC_FREQ;
  header.byte_p_sec = ((bits > 8) ? 2 : 1) * header.modus * DEC_FREQ;
  header.byte_p_spl = (bits > 8) ? 2 : 1;
  header.bit_p_spl = bits;

  fast_memcpy (&(header.data_chunk), data, 4);
  header.data_length = data_length;
  write (sfdes, &header, sizeof (header));
}


int
open_sound_file (int type)
{
  int themask, sfdes;
  char filename[100] = "";
  char num[4] = "";
  char aux[100] = "";
  char ext[5];
  int i;

  if (type)
    strcpy (ext, ".wav");
  else
    strcpy (ext, ".raw");

  strcpy (filename, DEFAULT_FILE_PATH);
  strcat (filename, "/nagraudio");

  for (i = 0; i < 1000; i++)
    {
      sprintf (num, "%.3d", i);
      strcpy (aux, filename);
      strcat (aux, num);
      strcat (aux, ext);
      if (open (aux, O_RDONLY) == -1)
        break;
    }

  themask = umask (0);
  umask (themask);
  sfdes = open (aux, O_CREAT | O_TRUNC | O_WRONLY, (~themask) & 0666);

  if (sfdes == -1)
    fprintf (stderr, "Error opening sound file\n");
  else
    {
      printf ("Writing to %s ...\n", aux);
      if (type)
        write_wav_header (sfdes, 0);
    }

  return sfdes;

}

void
close_sound_file (int sfdes, int type, int writed_buf)
{
  if (type == 2)
    {
      lseek (sfdes, 0, SEEK_SET);
      write_wav_header (sfdes, writed_buf * SIZE_BUFFER * sizeof (short int));
    };
  close (sfdes);
}
#endif // WRITE_TO_WAV

void
error_exit (char *message, int number)
{
  fprintf (stderr, "\n");
  fprintf (stderr, message);
  fprintf (stderr, "\nExiting Nagraudio thread.\n");
  pthread_exit ((void *) number);
}

void
do_audio_decode_44 ()
{

  short int *pre_buffer_in, *aux1;
  short int *pre_buffer_out, *aux2;
  short int *post_buffer_in, *aux3;
  short int *post_buffer_out, *aux4;
  short int *modulation_in, *modulation_out, *aux5;
  short int *dec_table, *out;
  int i;
  int j, temp_out;                /* Filters variables */

  int k, pos, changes;            /* Auto decode */
  int inertia = 0;                /* variables */

#ifdef WRITE_TO_WAV
  int sfdes, writed_buf, writing = 0;    /* Sound file descriptor & writing flag */
#endif

#ifdef ALSA
  ssize_t r, w;
  char aux_string[100];
#endif

  struct timeval tv;

  pre_buffer_in = malloc ((SIZE_BUFFER + MAX_ORDER) * sizeof (short int));
  memset (pre_buffer_in, 0, (SIZE_BUFFER + MAX_ORDER) * sizeof (short int));
  pre_buffer_in += MAX_ORDER;
  aux1 = pre_buffer_in;

  pre_buffer_out = malloc ((SIZE_BUFFER + MAX_ORDER) * sizeof (short int));
  memset (pre_buffer_out, 0, (SIZE_BUFFER + MAX_ORDER) * sizeof (short int));
  pre_buffer_out += MAX_ORDER;
  aux2 = pre_buffer_out;

  post_buffer_in = malloc ((SIZE_BUFFER + MAX_ORDER) * sizeof (short int));
  memset (post_buffer_in, 0, (SIZE_BUFFER + MAX_ORDER) * sizeof (short int));
  post_buffer_in += MAX_ORDER;
  aux3 = post_buffer_in;

  post_buffer_out = malloc ((SIZE_BUFFER + MAX_ORDER) * sizeof (short int));
  memset (post_buffer_out, 0, (SIZE_BUFFER + MAX_ORDER) * sizeof (short int));
  post_buffer_out += MAX_ORDER;
  aux4 = post_buffer_out;

  dec_table = malloc (SIZE_BUFFER * sizeof (short int));
  aux5 = dec_table;

  for (i = 0; i < SIZE_BUFFER; i++)
    *dec_table++ =
      (short
       int) ((cos
              (M_PI * (float) i *
               (2. * (float) NAGRA_FREQ / (float) DEC_FREQ)) *
              (1 << PRECISSION)) - 1);

#ifdef ALSA
  prepare_card_alsa ();
#else
  prepare_card_oss (16);
#endif

  while (1)
    {

#ifdef ALSA
      if ((r = snd_pcm_read (handle, aux1, SIZE_BUFFER * sizeof (short int)))
          < 0)
        {
          sprintf (aux_string, "Error reading device: %s\n",
                   snd_strerror (r));
          error_exit (aux_string, 2);
        }
#else
      if (read (fd, aux1, SIZE_BUFFER * sizeof (short int)) == -1)
        error_exit ("Error reading device", 2);
#endif

#if defined(HAVE_DIVX4LINUX) || defined (HAVE_FFMPEG)  || defined (HAVE_XVID)
      gettimeofday (&tv, NULL);
      audiostamp=tv.tv_usec/1000000.0+tv.tv_sec;
#endif

      /******************* AUTODECODE RUTINE *******************/
      if (auto_decode)
        {
          changes = 0;
          pos = *(aux1) > 0;
          for (k = 1; k < 20; k++)
            if (*(aux1 + k) > 10)
              {
                if (!pos)
                  {
                    changes++;
                    pos = 1;
                  }
              }
            else if (*(aux1 + k) < -10)
              if (pos)
                {
                  changes++;
                  pos = 0;
                }


          if (changes > 6)
            {
              if (inertia < 5)
                inertia++;
            }
          else if (inertia > -5)
            inertia--;

          decode = !(inertia < 0);
        }
      /************* END OF AUTODECODE RUTINE ****************/

      if (decode)
        {
      /**************** PRE MODULATION FILTER ******************/
          if (pre_filter_level)
            {
              pre_buffer_in = aux1;
              pre_buffer_out = aux2;

              for (i = 0; i < SIZE_BUFFER; i++)
                {
                  temp_out = 0;

                  for (j = 1; j < pre_filter_level + 1; j++)
                    temp_out +=
                      pre_out_coef[pre_filter_level][j] * *(pre_buffer_out -
                                                            j);

                  temp_out = temp_out >> PRECISSION;

                  for (j = 1; j < pre_filter_level / 2 + 1; j++)
                    temp_out +=
                      in_coef[pre_filter_level][j] * (*(pre_buffer_in - j) +
                                                      *(pre_buffer_in -
                                                        pre_filter_level +
                                                        j));

                  *pre_buffer_in =
                    (*pre_buffer_in << PRECISSION) /
                    pre_gains[pre_filter_level];

                  temp_out +=
                    *(pre_buffer_in - pre_filter_level) + *pre_buffer_in;

                  pre_buffer_in++;
                  *pre_buffer_out++ = temp_out;
                }

              modulation_in = aux2;

              fast_memcpy (aux2 - MAX_ORDER, aux2 + SIZE_BUFFER - MAX_ORDER,
                      MAX_ORDER * sizeof (short int));
              fast_memcpy (aux1 - MAX_ORDER, aux1 + SIZE_BUFFER - MAX_ORDER,
                      MAX_ORDER * sizeof (short int));

            }
          else
            modulation_in = aux1;
      /************** END OF PRE MODULATION FILTER **************/

      /************** BEGIN OF MODULATION ******************/

          dec_table = aux5;
          modulation_out = aux3;

          /*
          if (have_MMX)
            {                    // use MMX

              pxor_r2r (mm7, mm7);
              movd_m2r (OUT_GAIN, mm7);

              for (i = 0; i < SIZE_BUFFER; i += 4)
                {

                  movq_m2r (*modulation_in, mm0);    // data
                  movq_m2r (*dec_table, mm1);

                  movq_r2r (mm0, mm2);
                  pmullw_r2r (mm1, mm0);
                  pmulhw_r2r (mm2, mm1);

                  movq_r2r (mm0, mm2);
                  punpcklwd_r2r (mm1, mm0);
                  punpckhwd_r2r (mm1, mm2);
                  psrad_i2r (PRECISSION, mm0);
                  psrad_i2r (PRECISSION, mm2);
                  pslld_r2r (mm7, mm0);
                  pslld_r2r (mm7, mm2);

                  packssdw_r2r (mm2, mm0);
                  movq_r2m (mm0, *modulation_out);

                  modulation_out += 4;
                  modulation_in += 4;
                  dec_table += 4;
                }
              emms ();

            }
          else
*/
            for (i = 0; i < SIZE_BUFFER; i++)
              *modulation_out++ =
                (*modulation_in++ ** dec_table++) >> (PRECISSION - OUT_GAIN);

      /************** END OF MODULATION ********************/

      /******************* POST MODULATION FILTER *****************/
          if (post_filter_level)
            {
              post_buffer_in = aux3;
              post_buffer_out = aux4;

              for (i = 0; i < SIZE_BUFFER; i++)
                {
                  temp_out = 0;

                  for (j = 1; j < post_filter_level + 1; j++)
                    temp_out +=
                      post_out_coef[post_filter_level][j] *
                      *(post_buffer_out - j);

                  temp_out = temp_out >> PRECISSION;

                  for (j = 1; j < post_filter_level / 2 + 1; j++)
                    temp_out +=
                      in_coef[post_filter_level][j] * (*(post_buffer_in - j) +
                                                       *(post_buffer_in -
                                                         post_filter_level +
                                                         j));

                  *post_buffer_in =
                    (*post_buffer_in << PRECISSION) /
                    post_gains[post_filter_level];

                  temp_out +=
                    *(post_buffer_in - post_filter_level) + *post_buffer_in;

                  post_buffer_in++;
                  *post_buffer_out++ = temp_out;
                }

              out = aux4;

              fast_memcpy (aux3 - MAX_ORDER, aux3 + SIZE_BUFFER - MAX_ORDER,
                      MAX_ORDER * sizeof (short int));
              fast_memcpy (aux4 - MAX_ORDER, aux4 + SIZE_BUFFER - MAX_ORDER,
                      MAX_ORDER * sizeof (short int));

            }
          else
            out = aux3;
      /*************** END OF POST MODULATION FILTER **************/

        }
      else
        {
          fast_memcpy (aux4, aux1, SIZE_BUFFER * sizeof (short int));
          out = aux4;
        }

#ifdef ALSA
      if ((w = snd_pcm_write (handle, out, SIZE_BUFFER * sizeof (short int)))
          < 0)
        {
          sprintf (aux_string, "Error writing device: %s (%d)\n",
                   snd_strerror (w), w);
          error_exit (aux_string, 2);
        }
#else
      if (write (fd, out, SIZE_BUFFER * sizeof (short int)) == -1)
        error_exit ("Error writing device", 2);
#endif

#if defined(HAVE_DIVX4LINUX) || defined (HAVE_FFMPEG)  || defined (HAVE_XVID)
      pthread_mutex_lock(&divx_mutex);
      if (is_divx_in_progress())
        write_audio_to_avi(out, SIZE_BUFFER);
      pthread_mutex_unlock(&divx_mutex);
#endif

#ifdef WRITE_TO_WAV
      if (writing)
        {
          if (write (sfdes, out, SIZE_BUFFER * sizeof (short int)) == -1)
            {
              error_exit ("Error writing file", 2);
            }
          else
            {
              writed_buf++;
            }
        }
#endif

      pthread_testcancel ();
    }

}


#ifndef ALSA
void
do_audio_decode_8 ()
{
  unsigned char buffer[SIZE_BUFFER];
  unsigned char dec_table[256];
  int i;

  for (i = 0; i < 255; i++)
    dec_table[i] = 255 - i;

  prepare_card_oss (8);

  while (1)
    {
      if (read (fd, buffer, SIZE_BUFFER) == -1)
        perror ("read");

      for (i = 0; i < SIZE_BUFFER; i += 2)
        buffer[i] = dec_table[buffer[i]];

      if (write (fd, buffer, SIZE_BUFFER) == -1)
        perror ("write");

      pthread_testcancel ();
    }
}
#endif

#ifdef ALSA
int
prepare_card_alsa ()
{
  snd_pcm_format_t format;
  snd_pcm_channel_params_t playback_params;
  snd_pcm_channel_params_t capture_params;
  int card = 0;
  int device = 0;
  int err;

  char aux_string[100];

  if (debug)
    fprintf(stderr, "prepare_card_alsa\n");

  if ((card = snd_defaults_pcm_card ()) < 0)
    {
      sprintf (aux_string, "There is no default pcm card.\n");
      error_exit (aux_string, -1);
    }

  if ((device = snd_defaults_pcm_device ()) < 0)
    {
      sprintf (aux_string, "There is no default pcm device.\n");
      error_exit (aux_string, -1);
    }


  if ((err = snd_pcm_open (&handle, card, device, SND_PCM_OPEN_DUPLEX)) < 0)
    {
      sprintf (aux_string, "open failed: %s\n", snd_strerror (err));
      error_exit (aux_string, -1);
    }

  /* Set FORMAT */
  memset (&format, 0, sizeof (snd_pcm_format_t));
  format.interleave = 1;
  format.format = FORMAT;
  format.rate = DEC_FREQ;
  format.voices = MONO_STEREO;

  /* Set PLAYBACK parameters */
  snd_pcm_channel_flush (handle, SND_PCM_CHANNEL_PLAYBACK);
  memset (&playback_params, 0, sizeof (snd_pcm_channel_params_t));
  playback_params.channel = SND_PCM_CHANNEL_PLAYBACK;
  playback_params.mode = SND_PCM_MODE_BLOCK;
  fast_memcpy (&playback_params.format, &format, sizeof (format));
  playback_params.start_mode = SND_PCM_START_FULL;
  //  playback_params.stop_mode = SND_PCM_STOP_ROLLOVER;
  playback_params.stop_mode = SND_PCM_STOP_STOP;
  playback_params.buf.block.frag_size = FRAGMENT_S;
  playback_params.buf.block.frags_min = 1;
  playback_params.buf.block.frags_max = NUMBER_FRAG;
  playback_params.buf.block.frags_max = -1;

  if ((err = snd_pcm_channel_params (handle, &playback_params)) < 0)
    {
      sprintf (aux_string, "playback parameters setup failed: %s\n",
               snd_strerror (err));
      error_exit (aux_string, -1);
    }

  if ((err = snd_pcm_channel_prepare (handle, SND_PCM_CHANNEL_PLAYBACK)) < 0)
    {
      sprintf (aux_string, "playback prepare failed: %s\n",
               snd_strerror (err));
      error_exit (aux_string, -1);
    }

  /*
     if ( (err = snd_pcm_channel_go( handle, SND_PCM_CHANNEL_PLAYBACK )) < 0 ) {
     sprintf(aux_string, "Initiating playback transfer failed: %s\n",snd_strerror(err));
     error_exit(aux_string,-1);
     }
   */

  /* Set CAPTURE parameters */
  snd_pcm_channel_flush (handle, SND_PCM_CHANNEL_CAPTURE);
  memset (&capture_params, 0, sizeof (snd_pcm_channel_params_t));
  capture_params.channel = SND_PCM_CHANNEL_CAPTURE;
  capture_params.mode = SND_PCM_MODE_BLOCK;
  fast_memcpy (&capture_params.format, &format, sizeof (format));
  capture_params.start_mode = SND_PCM_START_DATA;
  capture_params.stop_mode = SND_PCM_STOP_ROLLOVER;
  //capture_params.stop_mode = SND_PCM_STOP_STOP;
  capture_params.buf.block.frag_size = FRAGMENT_S;
  capture_params.buf.block.frags_min = 1;
  capture_params.buf.block.frags_max = -1;
  //capture_params.buf.block.frags_max = 1;
  //capture_params.buf.block.frags_max = NUMBER_FRAG;

  if ((err = snd_pcm_channel_params (handle, &capture_params)) < 0)
    {
      sprintf (aux_string, "capture params setup failed: %s\n",
               snd_strerror (err));
      error_exit (aux_string, -1);
    }

  if ((err = snd_pcm_channel_prepare (handle, SND_PCM_CHANNEL_CAPTURE)) < 0)
    {
      sprintf (aux_string, "capture prepare failed: %s\n",
               snd_strerror (err));
      error_exit (aux_string, -1);
    }


  /*
     if ( (err = snd_pcm_channel_go( handle, SND_PCM_CHANNEL_CAPTURE )) < 0 ) {
     sprintf(aux_string, "Initiating capture transfer failed: %s\n",snd_strerror(err));
     error_exit(aux_string,-1);
     }
   */

  return 0;
}


#else

int
prepare_card_oss (int bits)
{
  int aux, caps;

  if (debug)
    fprintf(stderr, "prepare_card_oss\n");

  fd = open (device, O_RDWR);
  if (fd < 0)
    error_exit ("Error opening audio device", -1);

  if (ioctl (fd, SNDCTL_DSP_SETDUPLEX, 0) == -1)
    error_exit ("Error with full duplex setup\n", -1);

  if (ioctl (fd, SNDCTL_DSP_GETCAPS, &caps) == -1)
    error_exit ("Sorry but your sound driver is too old\n", -1);

  if (!(caps & DSP_CAP_DUPLEX))
    error_exit ("Sorry but your soundcard can't do full duplex\n", -1);

  aux = FRAGMENTS;
  if (ioctl (fd, SNDCTL_DSP_SETFRAGMENT, &aux) == -1)
    error_exit ("Error setting audio fragments", -1);

  if (bits == 8)
    aux = FORMAT8;
  else
    aux = FORMAT16;

  if (ioctl (fd, SNDCTL_DSP_SETFMT, &aux) == -1)
    error_exit ("Error setting audio format (bits)", -1);

  aux = MONO_STEREO;
  if (ioctl (fd, SNDCTL_DSP_STEREO, &aux) == -1)
    error_exit ("Error setting audio format (stereo)", -1);

  aux = DEC_FREQ;
  if (ioctl (fd, SNDCTL_DSP_SPEED, &aux) == -1)
    error_exit ("Error setting audio format (frequency)", -1);

  return 0;
}

#endif

void
close_card (void)
{
#ifdef ALSA
  snd_pcm_channel_flush (handle, SND_PCM_CHANNEL_PLAYBACK);
  snd_pcm_channel_flush (handle, SND_PCM_CHANNEL_CAPTURE);
  snd_pcm_close (handle);
  handle = NULL;
#else
  close (fd);
  fd = -1;
#endif
}


void
read_nagraudio_config (void)
{
  FILE *fp;
  char filename[100], line[100], tag[32], val[100], *h;
  int nr = 0;

  sprintf (filename, "%s/%s", getenv ("HOME"), ".nagraudiorc");

  fp = fopen (filename, "r");

  if (NULL == fp)
    fp = fopen ("/etc/nagraudiorc", "r");

  if (NULL == fp)
    {
      fprintf (stderr, "Can't open config file, using defaults\n\n");
      return;
    }

  while (NULL != fgets (line, 99, fp))
    {
      nr++;
      if (line[0] == '\n' || line[0] == '#' || line[0] == '%')
        continue;
      if (2 != sscanf (line, " %31[^= ] = %99[^\n]", tag, val))
        {
          fprintf (stderr, "%s:%d: parse error\n", filename, nr);
          continue;
        }

      for (h = val + strlen (val); h > val && *h == ' '; h--)
        *h = '\0';

      parse_nagraudio_config_line (tag, val, filename, nr);

    }
}

void
parse_nagraudio_config_line (char *tag, char *val, char *filename, int nr)
{
  //printf("In file %s at line %d: parsing Nagraudio parameter: %s = %s\n", filename, nr, tag, val);
  if (0 == strcmp (tag, "pre_filter"))
    if (atoi (val) != 0)
      if (atoi (val) < 1 || atoi (val) > MAX_ORDER / 2 + 1)
        fprintf (stderr, "%s:%d: wrong valor in %s\n", filename, nr, tag);
      else
        pre_filter_level = (atoi (val)) * 2 - 1;
    else
      pre_filter_level = 0;

  else if (0 == strcmp (tag, "post_filter"))
    if (atoi (val) != 0)
      if (atoi (val) < 1 || atoi (val) > MAX_ORDER / 2 + 1)
        fprintf (stderr, "%s:%d: wrong valor in %s\n", filename, nr, tag);
      else
        post_filter_level = (atoi (val)) * 2 - 1;
    else
      post_filter_level = 0;

  else if (0 == strcmp (tag, "output_gain"))
    if (atoi (val) != 0)
      if (atoi (val) < 1 || atoi (val) > MAX_GAIN)
        fprintf (stderr, "%s:%d: wrong valor in %s\n", filename, nr, tag);
      else
        OUT_GAIN = atoi (val);
    else
      OUT_GAIN = 0;

#ifndef ALSA
  else if (0 == strcmp (tag, "device"))
    strcpy (device, val);
#endif

  else if (0 == strcmp (tag, "auto_decode"))
    if (atoi (val) == 0)
      auto_decode = 0;
    else if (atoi (val) == 1)
      auto_decode = 1;
    else
      fprintf (stderr, "%s:%d: wrong valor in %s\n", filename, nr, tag);

  /*
     else if (0 == strcmp(tag,"gui"))
     if (atoi(val)==0)
     gui=0;
     else if (atoi(val)==1)
     gui=1;
     else
     fprintf(stderr,"%s:%d: wrong valor in %s\n",filename,nr,tag);
   */

  else if (0 == strcmp (tag, "8_bits"))
    if (atoi (val) == 0)
      bits = 16;
    else if (atoi (val) == 1)
      bits = 8;
    else
      fprintf (stderr, "%s:%d: wrong valor in %s\n", filename, nr, tag);

  else
    {
      fprintf (stderr, "%s:%d: unknown option %s\n", filename, nr, tag);
    }
}

void
save_nagraudio_config (FILE * fp)
{
  char *prefix = "nagraudio.";
  int val;

  if (fp == (FILE *) NULL)
    return;

  fprintf (fp, "#\n");
  fprintf (fp, "# Nagraudio configuration file\n");
  fprintf (fp, "# see Readme for more details\n");
  fprintf (fp, "#\n");
  fprintf (fp, "# 1: Decode in 8 bits mode ; 0: Use 16 bits mode\n");
  fprintf (fp, "%s8_bits = %d\n\n", prefix, (bits == 16) ? 0 : 1);

#ifndef ALSA
  fprintf (fp, "# Audio device to use\n");
  fprintf (fp, "%sdevice = %s\n\n", prefix, device);
#endif

  fprintf (fp, "# 1: Automatic decoding ; 0: Always decoding\n");
  fprintf (fp, "%sauto_decode = %d\n\n", prefix, auto_decode);

  fprintf (fp, "# 0: Disabled ; [1..4]: Level of filter to use\n");
  if (pre_filter_level == 0)
    val = 0;
  else
    val = (pre_filter_level + 1) / 2;
  fprintf (fp, "%spre_filter = %d\n\n", prefix, val);

  fprintf (fp, "# 0: Disabled ; [1..4]: Level of filter to use\n");
  if (post_filter_level == 0)
    val = 0;
  else
    val = (post_filter_level + 1) / 2;
  fprintf (fp, "%spost_filter = %d\n\n", prefix, val);

  fprintf (fp, "# 0: Disabled; 1: Output x2 ; 2: Output x4 ; 3: Output x8\n");
  fprintf (fp, "%soutput_gain = %d\n\n", prefix, OUT_GAIN);

}


void *
nagraudio_main (void *arg)
{
  int use_config = (int) arg;

  //printf("nagraudio_main: parameter value is: %d\n", use_config);

  pthread_setcancelstate (PTHREAD_CANCEL_ENABLE, NULL);

  if (use_config)
    read_nagraudio_config ();

  if (bits == 8)
    {
      pre_filter_level = 0;
      post_filter_level = 0;
      OUT_GAIN = 0;
      auto_decode = 0;
    }

  //printf("Nagraudio version %s\n", NAGRAUDIO_VERSION);
  if (debug)
    {
      printf ("Nagraudio version %s\n", NAGRAUDIO_VERSION);
      printf ("Sampling %d hz %d bits\n\n", DEC_FREQ, bits);

      if (!use_config)
        printf ("Ignoring configuration file\n");

      if (post_filter_level)
        printf ("Using post filter of level %d\n", post_filter_level / 2 + 1);

      if (pre_filter_level)
        printf ("Using pre filter of level %d\n", pre_filter_level / 2 + 1);

      if (auto_decode)
        printf ("Automatic decoding enabled.\n");

      if (OUT_GAIN)
        printf ("Using output gain of level %d (x%d)\n", OUT_GAIN,
                1 << OUT_GAIN);
    }

#ifndef ALSA
  if (bits == 8)
    do_audio_decode_8 ();
  else
    do_audio_decode_44 ();
#else
  do_audio_decode_44 ();
#endif

  return NULL;

}

void
set_nagraudio_auto_mode (int j)
{
  if ((j >= 0) && (j <= 1))
    auto_decode = j;
}

int
get_nagraudio_auto_mode (void)
{
  return auto_decode;
}

void
set_nagraudio_pre_filter (int j)
{
  if ((j >= 0) && (j <= 4))
    {
      if (j == 0)
        pre_filter_level = 0;
      else
        pre_filter_level = j * 2 - 1;
    }
}

int
get_nagraudio_pre_filter (void)
{
  return (pre_filter_level + 1) / 2;
}

void
set_nagraudio_post_filter (int j)
{
  if ((j >= 0) && (j <= 4))
    {
      if (j == 0)
        post_filter_level = 0;
      else
        post_filter_level = j * 2 - 1;
    }
}

int
get_nagraudio_post_filter (void)
{
  return (post_filter_level + 1) / 2;
}

void
set_nagraudio_gain (int j)
{
  if ((j >= 0) && (j <= 3))
    OUT_GAIN = j;
}

int
get_nagraudio_gain (void)
{
  return OUT_GAIN;
}

void
set_nagraudio_bit_mode (int j)
{

  switch (j)
    {
    case 0:
      bits = 16;
      break;
    case 1:
      bits = 8;
      /*
         pre_filter_level=0;
         post_filter_level=0;
         OUT_GAIN=0;
         auto_decode=0;      
       */
      break;
    default:
      break;
    }

}

int
get_nagraudio_bit_mode (void)
{
  if (bits == 8)
    return 1;
  else
    return 0;
}

void
wait_write_audio(void)
{
#if defined(HAVE_DIVX4LINUX) || defined (HAVE_FFMPEG)  || defined (HAVE_XVID)
  pthread_mutex_lock(&divx_mutex);
  pthread_mutex_unlock(&divx_mutex);
#endif
}
